port module Main exposing (..)

import Types exposing (..)
import Utils.Api exposing (..)
import Utils.HelperFunction exposing (..)

import Browser
import Browser.Navigation as Nav

import Json.Decode
import Url
import Time

import View

-- Ports for communication with JavaScript for Firestore requests and responses.
port requestFromFirestore : String -> Cmd msg
port receiveFromFirestore : (Json.Decode.Value -> msg) -> Sub msg
port receiveFirestoreError : (String -> msg) -> Sub msg

-- Main entry point for the Elm application.
main : Program () Model Msg
main =
    Browser.application
        { init = init
        , view = view
        , update = update
        , subscriptions = subscriptions
        , onUrlChange = UrlChanged
        , onUrlRequest = LinkClicked
        }

-- Initialize the model and set up initial state.
init : () -> Url.Url -> Nav.Key -> ( Model, Cmd Msg )
init _ url key =
    ( { key = key
      , url = url 
      , route = getRoute url
      , error = ""
      , quizId = ""
      , quiz = initQuiz
      , curTime = 0
      , runTime = False }
    , Cmd.none )

-- Set up subscriptions to receive messages from Firestore and handle timer ticks.
subscriptions : Model -> Sub Msg
subscriptions model =
    if model.runTime then
        Time.every 60000 Tick
    else
        Sub.batch 
            [ receiveFromFirestore (Json.Decode.decodeValue partialQuizDecoder >> ReceivePartialQuiz)
            , receiveFirestoreError ReceiveError ]

-- Handle the messages and update the model accordingly.
update : Msg -> Model -> ( Model , Cmd Msg )
update msg model =
    case msg of
        -- Handle sending a quiz request to Firestore.
        SendQuizRequest ->
            ( model, requestFromFirestore model.quizId )

        -- Handle receiving a partial quiz from Firestore.
        ReceivePartialQuiz (Ok partialQuiz) ->
            let
                quiz = transformPartialQuiz partialQuiz
            in
            ( { model | quiz = quiz, curTime = 0 }
                    , Nav.pushUrl model.key "/question/-1" )

        ReceivePartialQuiz (Err _) ->
            Debug.log "Error getting data from Firebase!" 
            (model, Cmd.none)

        -- Handle receiving an error from Firestore.
        ReceiveError errorMessage ->
            ( { model | error = errorMessage, quizId = "" }
                    , Cmd.none )

        -- Handle URL navigation within the app.
        LinkClicked urlRequest ->
            case urlRequest of
                Browser.Internal url ->
                    ( model, Nav.pushUrl model.key (Url.toString url) )

                Browser.External href ->
                    ( model, Nav.load href )

        -- Handle URL changes and update the route in the model.
        UrlChanged url ->
            let
                newRoute = getRoute url
                startRunTime = case newRoute of
                    QuestionRoute _ -> True
                    _ -> False
            in
            ( { model | url = url, runTime = startRunTime, route = newRoute }
            , Cmd.none
            )

        -- Handle updating the quiz ID.
        UpdateQuizId id ->
            ( { model | quizId = id }, Cmd.none )

        -- Handle answering a question and updating the model with the new answer.
        AnswerQuestion qid answer ->
            let
                q = model.quiz
                newQuestions = updateQuestionsAt qid answer model.quiz.questions
                newQuiz = { q | questions = newQuestions }
            in
            ( { model | quiz = newQuiz }, Cmd.none )

        -- Handle timer tick and update the current time, pushing to result view if time is up.
        Tick _ -> 
            let
                newTime = model.curTime + 1
                cmd = case model.quiz.duration of
                    Just t ->
                        if t <= newTime then Nav.pushUrl model.key "/result"
                        else Cmd.none
                    _ -> Cmd.none
            in
            ( { model | curTime = newTime }, cmd)

        -- Handle navigation to a specific question.
        GoToQuestion n ->
            let
                cmd = Nav.pushUrl model.key ("/question/" ++ String.fromInt n)
            in
            ( model, cmd )

        -- Handle navigation to the next question.
        GoToNextQuestion ->
            let
                cmd = case model.route of
                    QuestionRoute j -> Nav.pushUrl model.key ("/question/" ++ String.fromInt (j + 1))
                    _ -> Cmd.none
            in
            ( model, cmd )

        -- Handle navigation to the previous question.
        GoToPrevQuestion ->
            let
                cmd = case model.route of
                    QuestionRoute j -> Nav.pushUrl model.key ("/question/" ++ String.fromInt (j - 1))
                    _ -> Cmd.none
            in
            ( model, cmd )

        -- Handle resetting the quiz ID.
        ResetQuizID ->
            ( { model | quizId = "" } , Cmd.none )

        -- Handle resetting the current time.
        ResetCurTime ->
            ( { model | curTime = 0 }, Cmd.none )

-- Render the view based on the current model state.
view : Model -> Browser.Document Msg
view model =
    View.view model
