module Pages.QuizView exposing (questionView)

import Types exposing ( Msg(..), Quiz, Question, QuestionType(..))
import Html exposing (Html, div, text, h1, h2, h3, p, a, img, input, button)
import Html.Attributes exposing (src, class, href, placeholder, value)
import Html.Events exposing (onClick, onInput)
import Utils.SVGGenerator exposing (generateSVG)
import Utils.HelperFunction exposing (getQuestionAt)
import LaTeX.Expression

-- Function to render the view for a specific question in the quiz.
questionView : Int -> Quiz -> Html Msg
questionView i q =
    let
        -- Determine what to display based on the question index.
        toDisplay = if i < 0
            then
                displayFrontPage q
            else
                case getQuestionAt i q.questions of
                    Just question ->
                        displayQuestionPage question
                    _ ->
                        displayEndPage
        -- Navigation bar for moving between questions.
        navBarDiv = navBar i (List.length q.questions)
    in
    -- Combine the navigation bar with the current display content.
    div [ class "questionView" ]
        [ navBarDiv, toDisplay ]
    
-- Function to display the front page of the quiz, showing the quiz title and duration.
displayFrontPage : Quiz -> Html Msg
displayFrontPage q =
    let
        -- Display the quiz duration if it is specified.
        durationText = case q.duration of
            Just t -> p [] [ text ("Bearbeitungszeit: " ++ (String.fromInt t) ++ " Minuten")]
            Nothing -> p [] []
    in
    div [ class "quiz" ] 
        [ div [ class "frontPage" ]
            [ h1 [ class "title" ] [ text q.name ]
            , durationText ]
        , div [ class "navigation" ]
            [ nextPageBtn ] ]

-- Function to display the end page when the quiz is completed.
displayEndPage : Html Msg
displayEndPage =
    div [ class "quiz endPage" ] 
        [ h2 [] [ text "Das waren alle Fragen" ]
        , div [ class "navigation" ]
            [ prevPageBtn, finishQuizBtn ]]

-- Function to display a specific question page, including the navigation buttons.
displayQuestionPage : Question -> Html Msg
displayQuestionPage q = 
    div [ class "quiz" ]
        [ displayQuestion q 
        , div [ class "navigation"]
            [ prevPageBtn, nextPageBtn ]]

-- Function to display the content of a specific question, including hints, images, graphs, and answer options.
displayQuestion : Question  -> Html Msg
displayQuestion q =
    let
        -- Display the hint if available.
        hintDiv = case q.hint of
            Just hint -> LaTeX.Expression.compile hint
            _ -> div [ class "invisible" ] []
        -- Display the image if available.
        imageDiv = case q.image of
            Just url -> div [ class "imgDiv" ] [ img [ src url ] [ ] ]
            Nothing ->  div [ class "invisible" ] []
        -- Display the graph if available.
        funcDiv = case q.graph of
            Just gr -> div [ class "svgDiv" ] [ generateSVG {functions =  gr.functions, window = gr.window} ]
            Nothing -> div [ class "invisible" ] []
        -- Display the appropriate answer options based on the question type.
        answerDiv = case q.questionType of
            MCQ x -> displayMCQ q.id x
            SCQ x -> displaySCQ q.id x
            Input x -> displayInput q.id x
        -- Create a header prefix for the question.
        headerPrefix = "Frage " ++ String.fromInt (q.id + 1) ++ ":"
    in
    -- Combine all parts to display the question.
    div [ class "question" ] 
        [ div [ class "header" ] 
            [ h3 [ class "header-prefix" ] [ text headerPrefix ]
            , div [ class "header-task" ] [ LaTeX.Expression.compile q.task ]
            ]
        , hintDiv
        , div [ class "media" ] 
            [ imageDiv
            , funcDiv ]
        , answerDiv ]

-- Function to display Multiple Choice Questions (MCQ).
displayMCQ : Int -> { choices : List String, correct : List String, userSelection : List String } -> Html Msg
displayMCQ qid q =
    let
        -- Get the user's current selections.
        userAnswers = q.userSelection
        -- Create a div for each choice with LaTeX support and selection highlighting.
        divs = List.map (\choice ->
                let
                    selectedClass = if List.member choice userAnswers then "selected" else ""
                    latex = LaTeX.Expression.compile choice 
                in
                div [ class ("choice " ++ selectedClass), onClick (AnswerQuestion qid choice) ] 
                    [ latex ]
            ) q.choices
    in
    -- Wrap all choice divs in a container.
    div [ class "choicesMCQ" ] divs

-- Function to display Single Choice Questions (SCQ).
displaySCQ : Int -> { choices : List String, correct : String, userSelection : String } -> Html Msg
displaySCQ qid q =
    let
        -- Create a div for each choice with LaTeX support and selection highlighting.
        divs = List.map (\choice ->
                let
                    selectedClass = if choice == q.userSelection then "selected" else ""
                    latex = LaTeX.Expression.compile choice 
                in
                div [ class ("choice " ++ selectedClass), onClick (AnswerQuestion qid choice) ] 
                    [ latex ]
            ) q.choices
    in
    -- Wrap all choice divs in a container.
    div [ class "choicesSCQ" ] divs

-- Function to display input-based questions.
displayInput : Int -> { correct : List String, sensitive : Bool, userSelection : String } -> Html Msg
displayInput qid q =
    let
        -- Display the input field with the current user selection or placeholder text.
        inputToDisplay = if q.userSelection == "" 
            then input [ onInput (AnswerQuestion qid)
                , placeholder "Antwort"
                , class "inputField" ] [ ] 
            else input [ onInput (AnswerQuestion qid)
                , value q.userSelection
                , class "inputField" ] [ ] 
    in
    -- Wrap the input field in a container.
    div [ class "choiceInput" ] 
        [ inputToDisplay ]

-- Button to finish the quiz and go to the results page.
finishQuizBtn : Html Msg
finishQuizBtn =
    a [ href "/result", class "btn" ] [ text "Quiz beenden" ]

-- Button to go to the next question.
nextPageBtn : Html Msg
nextPageBtn =
    button 
        [ onClick GoToNextQuestion, class "navBtn" ]
        [ text "Weiter" ]

-- Button to go to the previous question.
prevPageBtn : Html Msg
prevPageBtn =
    button 
        [ onClick GoToPrevQuestion, class "navBtn" ] 
        [ text "Zurück" ]

-- Navigation bar for moving between questions.
navBar : Int -> Int -> Html Msg
navBar index n =
    let
        -- Create a button for each question in the quiz.
        buttons = List.range 0 (n - 1) |> List.map 
            (\x -> 
                let
                    classAttribute = if x == index then "navBarBtn curIndex" else "navBarBtn"
                in
                button 
                    [ onClick (GoToQuestion x), class classAttribute ] 
                    [ text ("Frage " ++ String.fromInt (x + 1))]
                )
    in
    -- Wrap the navigation buttons and finish button in a container.
    div [ class "navBar" ] 
        [ div [ class "navBarBtns" ] buttons
        , finishQuizBtn ]