module Pages.ResultView exposing (resultView)

import Types exposing (Model, Msg(..), Question, QuestionType(..))
import Html exposing (Html, div, text, button, a, p, h1, table, tbody, tr, td)
import Html.Attributes exposing (class, href)
import Html.Events exposing (onClick)
import Utils.HelperFunction exposing (normalizeString)
import LaTeX.Expression

-- Main function to render the result view of the quiz.
resultView : Model -> Html Msg
resultView model =
    let
        -- Convert the answers to an HTML table.
        content = answersToTable model.quiz.questions
    in
    div [ class "resultView" ]
        [ h1 [] [ text "Ergebnisse" ]
        , p [] [ text ("benötigt: " ++ String.fromInt model.curTime ++ " Minuten")]
        , content 
        , div [ class "navigation" ]
            [ button [ onClick SendQuizRequest, class "btn" ] [ text "wiederholen" ]
            , a [ class "btn", href "/home", onClick ResetQuizID ] [ text "neu" ]]
        ]

-- Convert list of questions to an HTML table.
answersToTable : List Question -> Html Msg
answersToTable q =
    table []
        [ tbody [] ( List.map answerToRow q) ]

-- Convert a single question to a table row.
answerToRow : Question -> Html Msg
answerToRow q =
    let
        key = "Frage " ++ (String.fromInt (q.id + 1))
        -- Check if the user's answer is correct.
        gotCorrect = evaluateQuestion q.questionType
        -- Create a colored div based on correctness.
        colorDiv = if gotCorrect 
            then 
                div [ class "colorDiv correct" ] []
            else 
                div [ class "colorDiv false" ] []
        -- Display the result based on question type.
        result = case q.questionType of
            MCQ a -> displayMcqResult gotCorrect a.correct a.userSelection
            SCQ a -> displayScqResult gotCorrect a.correct a.userSelection
            Input a -> displayInputResult gotCorrect a.correct a.userSelection
        task = div [ class "task" ] [ LaTeX.Expression.compile q.task ]
    in
    -- Wrap the key, color div, and result in a table row.
    tr []
        [ td [] [ text key ]
        , td [] [ colorDiv ]
        , td [] [ div [] [ task, result] ]
        ]

-- Evaluate if the user's answer to a question is correct.
evaluateQuestion : QuestionType -> Bool
evaluateQuestion q =
    case q of
        MCQ a -> areAllIn a.userSelection a.correct && areAllIn a.correct a.userSelection
        SCQ a -> a.userSelection == a.correct
        Input a -> 
            let
                str = if a.sensitive then a.userSelection else (normalizeString a.userSelection)
            in List.member str a.correct

-- Display result for Multiple Choice Questions (MCQ).
displayMcqResult : Bool -> List String -> List String -> Html Msg
displayMcqResult gotCorrect correctAnswers userAnswers = 
    let
        thisWasCorrect = String.join "  und  " correctAnswers
        yourAnswer = String.join "  und  " userAnswers
        feedback = if gotCorrect
            then 
                div [ class "innerResult" ]
                    [ div [ class "answer" ] 
                        [ p [ class "answer" ] [ text "Antwort: "]
                        , LaTeX.Expression.compile thisWasCorrect ] ]
            else 
                wrongAnswerDiv thisWasCorrect yourAnswer
    in
    -- Wrap the feedback in a result div.
    div [ class "result"] [ feedback ]

-- Display result for Single Choice Questions (SCQ).
displayScqResult : Bool -> String -> String -> Html Msg
displayScqResult gotCorrect correctAnswer userAnswer =
    let
        feedback = if gotCorrect
            then 
                div [ class "innerResult" ]
                    [ div [ class "answer" ] 
                        [ p [ class "answer" ] [ text "Antwort: " ]
                        , p [ class "follow" ] [ LaTeX.Expression.compile correctAnswer ] ]]
            else 
                wrongAnswerDiv correctAnswer userAnswer
    in
    -- Wrap the feedback in a result div.
    div [ class "result"] [ feedback ]

-- Display result for input-based questions.
displayInputResult : Bool -> List String -> String -> Html Msg
displayInputResult gotCorrect correctAnswers userAnswer =
    let
        thisWasCorrect = String.join "  oder  " correctAnswers
        feedback = if gotCorrect
            then 
                div [ class "innerResult" ]
                    [ div [ class "answer" ] 
                        [ p [ class "answer" ] [ text "Antwort: " ]
                        , p [ class "follow" ] [ LaTeX.Expression.compile thisWasCorrect ] ]]
            else 
                wrongAnswerDiv thisWasCorrect userAnswer
    in
    -- Wrap the feedback in a result div.
    div [ class "result"] [ feedback ]

-- Check if all elements in list 'a' are present in list 'b'.
areAllIn : List String -> List String -> Bool
areAllIn a b =
    List.all (\x -> List.member x a) b

-- Display feedback for incorrect answers.
wrongAnswerDiv : String -> String -> Html Msg
wrongAnswerDiv correctAnswer userAnswer =
    div [ class "innerResult" ] 
        [ div [ class "answer" ] 
            [ p [ class "answer" ] [ text "Richtig: " ] 
            , p [ class "follow" ] [ LaTeX.Expression.compile correctAnswer ]]
        , div [ class "vertical-line" ] []
        , div [ class "answer" ] 
            [ p [ class "answer" ] [ text "Deine Antwort: "]
            , p [ class "follow" ] [ LaTeX.Expression.compile userAnswer ] ]]
