module Types exposing (..)

import Browser
import Browser.Navigation as Nav
import Url
import Json.Decode
import Time

type alias Model =
    { key : Nav.Key
    , url : Url.Url
    , route : Route
    , error : String
    , quizId : String
    , quiz : Quiz
    , curTime : Int
    , runTime : Bool }

type Msg
    = SendQuizRequest       -- Message to send a quiz request
    | ReceivePartialQuiz (Result Json.Decode.Error PartialQuiz)  -- Message for receiving a partial quiz
    | ReceiveError String   -- Message for receiving an error
    | LinkClicked Browser.UrlRequest  -- Message for handling a clicked link
    | UrlChanged Url.Url    -- Message for handling a URL change
    | UpdateQuizId String   -- Message for updating the quiz ID
    | AnswerQuestion Int String  -- Message for answering a question
    | Tick Time.Posix       -- Message for ticking the timer
    | GoToQuestion Int      -- Message for navigating to a specific question
    | GoToNextQuestion      -- Message for navigating to the next question
    | GoToPrevQuestion      -- Message for navigating to the previous question
    | ResetQuizID           -- Message for resetting the quiz ID
    | ResetCurTime          -- Message for resetting the current time

-- The different routes in the application.
type Route
    = HomeRoute
    | QuestionRoute Int
    | ResultRoute
    | NotFound

-- Represents a graph, including its window and functions.
type alias Graph =
    { window : Window
    , functions : List MathFunction }


-- The window of a graph, specifying its boundaries.
type alias Window =
    { xMin : Int
    , yMin : Int
    , xMax : Int
    , yMax : Int
    }

-- The different types of mathematical functions.
type MathFunction
    = Const Float
    | X
    | Add MathFunction MathFunction
    | Sub MathFunction MathFunction
    | Mul MathFunction MathFunction
    | Div MathFunction MathFunction
    | Pow MathFunction MathFunction
    | Log MathFunction MathFunction
    | Ln  MathFunction
    | Neg MathFunction
    | Sin MathFunction
    | Cos MathFunction
    | Tan MathFunction
    | Exp MathFunction

-- Represents a quiz, including its ID, name, duration, and questions.
type alias Quiz =
    { id : String
    , name : String
    , duration : Maybe Int  -- Duration in minutes
    , questions : List Question
    }

-- Represents a single question in a quiz.
type alias Question =
    { id : Int
    , task : String
    , hint : Maybe String
    , explanation : Maybe String
    , image : Maybe String  -- URL to an image
    , graph : Maybe Graph
    , questionType : QuestionType
    }

-- The different types of questions that can be asked.
type QuestionType 
    = MCQ { choices : List String, correct : List String, userSelection : List String }
    | SCQ { choices : List String, correct : String, userSelection : String }
    | Input { correct : List String, sensitive : Bool, userSelection : String }

{-
 -
 - Representing partial Types, which we get from the server:
 -
 -}

type alias PartialQuiz =
    { id : String
    , name : String
    , duration : Maybe Int
    , questions : List PartialQuestion
    }

type alias PartialQuestion =
    { task : String 
    , hint : Maybe String
    , explanation : Maybe String
    , image : Maybe String 
    , graph : Maybe Graph 
    , questionType : PartialQuestionType }

type PartialQuestionType 
    = PartialMCQ { choices : List String, correct : List String }
    | PartialSCQ { choices : List String, correct : String }
    | PartialInput { correct : List String, sensitive : Bool }