module Utils.Api exposing (..)

import Types exposing (..)
import Json.Encode as E exposing (..)
import Json.Decode as D exposing (..)

encodeNullable : (value -> E.Value) -> Maybe value -> E.Value
encodeNullable valueEncoder maybeValue =
    case maybeValue of
        Just value ->
            valueEncoder value

        Nothing ->
            E.null

encodeGraph : Graph -> E.Value
encodeGraph graph =
    object 
        [ ("window", encodeWindow graph.window)
        , ("functions", E.list encodeMathFunction graph.functions)
        ]

encodeWindow : Window -> E.Value
encodeWindow window =
    object 
        [ ("xMin", E.int window.xMin)
        , ("xMax", E.int window.xMax)
        , ("yMin", E.int window.yMin)
        , ("yMax", E.int window.yMax)
        ]

encodeMathFunction : MathFunction -> E.Value
encodeMathFunction mathFunction =
    case mathFunction of
        Const x ->
            object [ ( "type", E.string "Const" ), ( "value", E.float x ) ]

        X ->
            object [ ( "type", E.string "X" ) ]

        Add f1 f2 ->
            object [ ( "type", E.string "Add" ), ( "left", encodeMathFunction f1 ), ( "right", encodeMathFunction f2 ) ]

        Sub f1 f2 ->
            object [ ( "type", E.string "Sub" ), ( "left", encodeMathFunction f1 ), ( "right", encodeMathFunction f2 ) ]

        Mul f1 f2 ->
            object [ ( "type", E.string "Mul" ), ( "left", encodeMathFunction f1 ), ( "right", encodeMathFunction f2 ) ]

        Div f1 f2 ->
            object [ ( "type", E.string "Div" ), ( "left", encodeMathFunction f1 ), ( "right", encodeMathFunction f2 ) ]

        Pow f1 f2 ->
            object [ ( "type", E.string "Pow" ), ( "left", encodeMathFunction f1 ), ( "right", encodeMathFunction f2 ) ]

        Log f1 f2 ->
            object [ ( "type", E.string "Log" ), ( "left", encodeMathFunction f1 ), ( "right", encodeMathFunction f2 ) ]

        Ln f ->
            object [ ( "type", E.string "Ln" ), ( "argument", encodeMathFunction f ) ]

        Neg f ->
            object [ ( "type", E.string "Neg" ), ( "argument", encodeMathFunction f ) ]

        Sin f ->
            object [ ( "type", E.string "Sin" ), ( "argument", encodeMathFunction f ) ]

        Cos f ->
            object [ ( "type", E.string "Cos" ), ( "argument", encodeMathFunction f ) ]

        Tan f ->
            object [ ( "type", E.string "Tan" ), ( "argument", encodeMathFunction f ) ]

        Exp f ->
            object [ ( "type", E.string "Exp" ), ( "argument", encodeMathFunction f ) ]


decodeMathFunctionInit : Decoder MathFunction
decodeMathFunctionInit =
    field "type" D.string |> andThen decodeMathFunction

decodeMathFunction : String -> Decoder MathFunction
decodeMathFunction type_ =
    case type_ of
        "Const" ->
            field "value" D.float |> map Const

        "X" ->
            D.succeed X

        "Add" ->
            map2 Add (field "left" decodeMathFunctionInit) (field "right" decodeMathFunctionInit)

        "Sub" ->
            map2 Sub (field "left" decodeMathFunctionInit) (field "right" decodeMathFunctionInit)

        "Mul" ->
            map2 Mul (field "left" decodeMathFunctionInit) (field "right" decodeMathFunctionInit)

        "Div" ->
            map2 Div (field "left" decodeMathFunctionInit) (field "right" decodeMathFunctionInit)

        "Pow" ->
            map2 Pow (field "left" decodeMathFunctionInit) (field "right" decodeMathFunctionInit)

        "Log" ->
            map2 Log (field "left" decodeMathFunctionInit) (field "right" decodeMathFunctionInit)

        "Ln" ->
            map Ln (field "argument" decodeMathFunctionInit)

        "Neg" ->
            map Neg (field "argument" decodeMathFunctionInit)

        "Sin" ->
            map Sin (field "argument" decodeMathFunctionInit)

        "Cos" ->
            map Cos (field "argument" decodeMathFunctionInit)

        "Tan" ->
            map Tan (field "argument" decodeMathFunctionInit)    

        "Exp" ->
            map Exp (field "argument" decodeMathFunctionInit)

        _ ->
            D.fail ("Unknown MathFunction type: " ++ type_)

decodeGraph : Decoder Graph
decodeGraph = 
    map2 Graph
        (field "window" decodeWindow)
        (field "functions" (D.list decodeMathFunctionInit))

decodeWindow : Decoder Window
decodeWindow =
    map4 Window
        (field "xMin" D.int)
        (field "yMin" D.int)
        (field "xMax" D.int)
        (field "yMax" D.int)

partialMCQDecoder : Decoder PartialQuestionType
partialMCQDecoder =
    D.map2 (\choices correct -> PartialMCQ { choices = choices, correct = correct })
        (D.field "choices" (D.list D.string))
        (D.field "correct" (D.list D.string))

partialSCQDecoder : Decoder PartialQuestionType
partialSCQDecoder =
    D.map2 (\choices correct -> PartialSCQ { choices = choices, correct = correct })
        (D.field "choices" (D.list D.string))
        (D.field "correct" D.string)

partialInputDecoder : Decoder PartialQuestionType
partialInputDecoder =
    D.map2 (\correct sensitive -> PartialInput { correct = correct, sensitive = sensitive })
        (D.field "correct" (D.list D.string))
        (D.field "sensitive" D.bool)

partialQuestionTypeDecoder : Decoder PartialQuestionType
partialQuestionTypeDecoder =
    D.oneOf
        [ D.field "MCQ" partialMCQDecoder
        , D.field "SCQ" partialSCQDecoder
        , D.field "Input" partialInputDecoder
        ]

partialQuestionDecoder : Decoder PartialQuestion
partialQuestionDecoder =
    D.map6 PartialQuestion
        (D.field "task" D.string)
        (D.field "hint" (D.maybe D.string))
        (D.field "explanation" (D.maybe D.string))
        (D.field "image" (D.maybe D.string))
        (D.field "graph" (D.maybe decodeGraph))
        (D.field "questionType" partialQuestionTypeDecoder)

partialQuizDecoder : Decoder PartialQuiz
partialQuizDecoder =
    D.map4 PartialQuiz
        (D.field "id" D.string)
        (D.field "name" D.string)
        (D.field "duration" (D.maybe D.int))
        (D.field "questions" (D.list partialQuestionDecoder))

partialQuestionTypeEncoder : PartialQuestionType -> E.Value
partialQuestionTypeEncoder questionType =
    case questionType of
        PartialMCQ { choices, correct } ->
            object
                [ ( "MCQ", 
                    object
                        [ ( "choices", E.list E.string choices )
                        , ( "correct", E.list E.string correct )
                        ]
                  )
                ]

        PartialSCQ { choices, correct } ->
            object
                [ ( "SCQ",
                    object
                        [ ( "choices", E.list E.string choices )
                        , ( "correct", E.string correct )
                        ]
                  )
                ]

        PartialInput { correct, sensitive } ->
            object
                [ ( "Input",
                    object
                        [ ( "correct", E.list E.string correct )
                        , ( "sensitive", E.bool sensitive )
                        ]
                  )
                ]

partialQuestionEncoder : PartialQuestion -> E.Value
partialQuestionEncoder question =
    object
        [ ( "task", E.string question.task )
        , ( "hint", encodeNullable E.string question.hint )
        , ( "explanation", encodeNullable E.string question.explanation )
        , ( "image", encodeNullable E.string question.image )
        , ( "graph", encodeNullable encodeGraph question.graph )
        , ( "questionType", partialQuestionTypeEncoder question.questionType )
        ]

partialQuizEncoder : PartialQuiz -> E.Value
partialQuizEncoder quiz =
    object
        [ ( "id", E.string quiz.id )
        , ( "name", E.string quiz.name )
        , ( "duration", encodeNullable E.int quiz.duration )
        , ( "questions", E.list partialQuestionEncoder quiz.questions )
        ]

transformPartialQuiz : PartialQuiz -> Quiz
transformPartialQuiz partialQuiz =
    let
        questions =
            List.indexedMap transformPartialQuestion partialQuiz.questions
    in
    { id = partialQuiz.id
    , name = partialQuiz.name
    , duration = partialQuiz.duration
    , questions = questions
    }

transformPartialQuestion : Int -> PartialQuestion -> Question
transformPartialQuestion index partialQuestion =
    { id = index
    , task = partialQuestion.task
    , hint = partialQuestion.hint
    , explanation = partialQuestion.explanation
    , image = partialQuestion.image
    , graph = partialQuestion.graph
    , questionType = transformPartialQuestionType partialQuestion.questionType
    }

transformPartialQuestionType : PartialQuestionType -> QuestionType
transformPartialQuestionType partialQuestionType =
    case partialQuestionType of
        PartialMCQ { choices, correct } ->
            MCQ { choices = choices, correct = correct, userSelection = [] }

        PartialSCQ { choices, correct } ->
            SCQ { choices = choices, correct = correct, userSelection = "" }

        PartialInput { correct, sensitive } ->
            Input { correct = correct, sensitive = sensitive, userSelection = "" }