module Utils.SVGGenerator exposing (generateSVG)

import Svg exposing (Svg, svg, path)
import Svg.Attributes exposing (d, viewBox, stroke, fill, strokeWidth, strokeOpacity, x, y, fontSize, textAnchor, dx, dy, points)
import Types exposing (Window, Graph)
import Utils.MathFunction exposing (toLambda)
import Svg.Attributes exposing (points)

-- Threshold value to detect discontinuities
threshold : Float
threshold = 10

-- Steps for discrete x-values
steps : Int
steps = 200

-- Main function to generate the SVG element
generateSVG : Graph -> Svg msg
generateSVG { functions, window } =
    let
        fns = List.map toLambda functions -- convert MathFunctions into Lambda functions
        viewBoxString = 
            String.join " " (List.map String.fromInt   [ window.xMin
                                                       , -window.yMax
                                                       , window.xMax - window.xMin
                                                       , window.yMax - window.yMin
                                                       ])
        -- Generate paths for each function
        paths = List.map (\fn -> generatePath fn window) fns
        
        -- Generate axis lines
        axis = 
            [ "M " ++ (String.fromInt window.xMin) ++ ",0 L " ++ (String.fromInt window.xMax) ++ ",0"
            , "M 0," ++ (String.fromInt -window.yMax) ++ " L 0," ++ (String.fromInt -window.yMin)
            ]
        
        -- Convert lists of paths to SVG path elements
        functionPaths = listToSVGPath paths "red" "0.75" "1.0"
        axisPaths = listToSVGPath axis "black" "0.5" "0.85"
        
        -- Generate grid lines
        gridPaths = generateGrid window "0.2"

        -- Axis labels
        xAxisLabel = svgText window.xMax 0 "x" "black" "0.5" "-0.4" "-0.1"
        yAxisLabel = svgText 0 -window.yMax "y" "black" "0.5" "0.2" "0.3"

        -- arrows at the end of each axis
        arrowLength = (toFloat (min (window.xMax - window.xMin) (window.yMax - window.yMin)))/20
        arrowWidth = String.fromFloat (arrowLength/4)
        xOffset = String.fromFloat (toFloat window.xMax - arrowLength)
        yOffset = String.fromFloat (toFloat -window.yMax + arrowLength)
        xArrow = Svg.polygon 
            [ points (String.fromInt window.xMax ++ ",0 " ++ xOffset ++ "," ++ arrowWidth ++ " " ++ xOffset ++ ",-" ++ arrowWidth)
            , stroke "black", strokeWidth "1%"] []
        yArrow = Svg.polygon 
            [ points ("0," ++ String.fromInt -window.yMax ++ " " ++ arrowWidth ++ "," ++ yOffset ++ " -" ++ arrowWidth ++ "," ++ yOffset)
            , stroke "black", strokeWidth "1%"] []
        
        -- Combine all SVG paths
        svgPaths = gridPaths ++ axisPaths ++ functionPaths ++ [xAxisLabel, yAxisLabel, xArrow, yArrow]
    in
    svg [ viewBox viewBoxString ] svgPaths

-- Function to generate a list of floats from a to b in n steps
range : Int -> Int -> Int -> List Float
range a b n =
    let
        step = (toFloat (b - a)) / (toFloat (n - 1))
    in
    List.map (\i -> toFloat a + step * toFloat i) (List.range 0 (n - 1))

-- Function to remove the last two characters from a string
removeLastTwoCharacters : String -> String
removeLastTwoCharacters str =
    String.left (String.length str - 2) str

-- Convert (x, y) coordinate to SVG path string
toPathString : (Float, Float) -> String
toPathString (x, y) =
    String.fromFloat x ++ "," ++ String.fromFloat -y

-- Generate path string for a given function and window
generatePath : (Float -> Float) -> Window -> String
generatePath fn viewBox =
    let
        values = range viewBox.xMin viewBox.xMax steps
        pathString = processValues values fn
    in
    pathString

-- Process values and generate SVG path string, handling discontinuities
processValues : List Float -> (Float -> Float) -> String
processValues values fn =
    let
        processHelper : List Float -> Maybe Float -> String -> String
        processHelper vals lastResult path =
            case vals of
                [] ->
                    removeLastTwoCharacters path

                value :: rest ->
                    let
                        result = fn value
                    in
                    if not (isInfinite result || isNaN result) then
                        case lastResult of
                            Just lastVal ->
                                if abs (result - lastVal) > threshold then
                                    processHelper rest (Just result) (removeLastTwoCharacters path ++ "M " ++ toPathString (value, result) ++ " L ")
                                else
                                    processHelper rest (Just result) (path ++ toPathString (value, result) ++ " L ")

                            Nothing ->
                                processHelper rest (Just result) (path ++ toPathString (value, result) ++ " L ")
                    else
                        processHelper rest Nothing (removeLastTwoCharacters path ++ "M ")
    in
    processHelper values Nothing "M "

-- Convert a list of path strings to a list of SVG path elements
listToSVGPath : List String -> String -> String -> String -> List (Svg msg)
listToSVGPath paths color weight opacity =
    List.map (\path -> stringToSVGPath path color weight opacity) paths

-- Convert a single path string to an SVG path element
stringToSVGPath : String -> String -> String -> String -> Svg msg
stringToSVGPath dStr color weight opacity =
    path [ d dStr, stroke color, strokeWidth (weight ++ "%"), fill "none", strokeOpacity opacity ] []

-- Generate grid lines for the given window
generateGrid : Window -> String -> List (Svg msg)
generateGrid window weight =
    let
        xMin = String.fromInt window.xMin
        xMax = String.fromInt window.xMax
        yMin = String.fromInt -window.yMin
        yMax = String.fromInt -window.yMax
        
        -- Generate vertical and horizontal grid lines
        verticalOffsets = List.range -window.yMax -window.yMin
        horizontalOffsets = List.range window.xMin window.xMax
        
        verticalLines = List.map (\y -> "M" ++ xMin ++ "," ++ y ++" L " ++ xMax ++ "," ++ y) (List.map String.fromInt verticalOffsets)
        horizontalLines = List.map (\x -> "M" ++ x ++ "," ++ yMin ++ " L " ++ x ++ "," ++ yMax) (List.map String.fromInt horizontalOffsets)
        
        verticalSVGPaths = listToSVGPath verticalLines "gray" weight "1.0"
        horizontalSVGPaths = listToSVGPath horizontalLines "gray" weight "1.0"
        xLabelsSVGPath = List.map (\x -> svgText x 0 (String.fromInt x) "black" "0.25" "0.15" "0.25") horizontalOffsets
        yLabelsSVGPath = List.map (\y -> svgText 0 y (String.fromInt -y) "black" "0.25" "-0.15" "-0.05") verticalOffsets
    in
    (verticalSVGPaths ++ horizontalSVGPaths ++ xLabelsSVGPath ++ yLabelsSVGPath)

-- Helper function to create SVG text elements
svgText : Int -> Int -> String -> String -> String -> String -> String -> Svg msg
svgText xPos yPos content color size dX dY =
    Svg.text_ 
        [ x (String.fromInt xPos)
        , y (String.fromInt yPos)
        , fill color
        , fontSize size
        , textAnchor "middle"
        , dx dX
        , dy dY ] 
        [ Svg.text content ]